/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 988 $ $LastChangedDate: 2002-09-15 12:23:51 +0200 (Sun, 15 Sep 2002) $
 *
 * Gont - Main compiler module, exports interface to be used by driver
 * program.
 *
 * Compilation phases are called on declaration list obtained from parser
 * seqentially. Current model is to first get all declartions from input
 * file, call phase1 on each declaration, call phase2 on each declaration,
 * and so on. 
 *
 * OTOH old compiler calls phase1 on first declaration, phase2 on first 
 * declaration, ..., phase1 on second declaration, ... and so on. This makes 
 * prototypes needed. 
 */
 
open Ast;
open Util;
open Xobjects;
open Gont;

int get_time()
{
	let (user, sys) = Sys::times() in
	return user + sys;
}

int last_time;

section init
{
	last_time = 0;
}

void print_time(string msg)
{
	if (flags.time_report) {
		int time = get_time();
		
		msg = msg + String::make(10 - String::length(msg), ' ');
		
		puts(msg + ": " + itoa(time - last_time) + "\n");
		
		last_time = time;
	}
}

<decl>list parse_and_bind(compilation_ctx ctx, bool iface)
{
	Bind::init_binder(ctx.modulename, iface);
	print_time("init");

	_ lex = Lexer::make_lexer(ctx.input_filename);
	
	<decl>list decls;
	
	try {
		decls = Parser::parse(fun () (Lexer::get_token(lex)));
	} with {
	case Failure("Parser::parser"):
		error("nothing parsed");
		raise Bail_out;
	}

	print_time("parse");
	
	void pass(string name, *(void (decl)) fnc)
	{
		List::iter(fnc, decls);
		print_time(name);
	}
	
	pass("enter", Bind::enter_decl);
	pass("bind", Bind::bind_decl);
	Bind::finish_bind();
	pass("tycheck", Type::tycheck_decl);
	pass("xtcheck", Xtcheck::xtcheck_decl);
	pass("chini", Chini::chini_decl);

	if (!iface) {
		Chini::chini_inits();
		print_time("chiniI");
	}

	maybe_choke();

	return decls;
}

Xml::node dump_type(decl d)
{
	switch d.a {
	case Tydef(s):
		Xml::node n = Xmldump::xmlize_ty_symbol(s);
		return n;
	case _:
		// ble.. broken by design
		<*(string, string)>list a = Nil;
		return Xml::Element("dummy", Nil, a);
	}
}

Stdio::FILE output_stream(compilation_ctx ctx)
{
	if (ctx.output_filename == "" || ctx.output_filename == "-")
		return Stdio::stdout;
	return Stdio::fopen(ctx.output_filename, "w");
}

void dump_xml_types(compilation_ctx ctx)
{
	let d = parse_and_bind(ctx, true) in
	let e = Xml::Element("tydefs", List::map(dump_type, d), Nil) in
	Xml::save_to_stream_indented(Xml::make_doc(e), output_stream(ctx));
}

void compile_interface(compilation_ctx ctx)
{
	parse_and_bind(ctx, true);
	Bind::save_module(ctx.output_filename);
	maybe_choke();
}

void compile_implementation(compilation_ctx ctx)
{
	<decl>list decls = parse_and_bind(ctx, false);
	Type::check_module_completness();
	maybe_choke();

	Stdio::FILE ksi_out = output_stream(ctx);
	Codegen::init(ksi_out);
	List::iter(Codegen::codegen_for_decl, decls);
	let (ctor, init, fini) = Codegen::fini() in {
		Stdio::fclose(ksi_out);

		if (ctx.output_xgo_filename != "") {
			xgo x = ({xgo_init_routine = init,
				  xgo_fini_routine = fini,
				  xgo_ctor_routine = ctor,
				  xgo_uses = Bind::used_modules(),
				  xgo_references = Bind::referenced_modules(),
				  xgo_used_exns = Bind::used_exns(),
				  xgo_name = ctx.modulename,
				  xgo_weak = false
				 });
			save_xgo(x, ctx.output_xgo_filename);
		}

		print_time("codegen");
	}
	
	last_time = 0;
	print_time("total");
	maybe_choke();
	
	if (flags.time_report)
		Sys::dump_allocator_statistics();
}
