/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 730 $ $LastChangedDate: 2002-08-31 09:23:01 +0200 (Sat, 31 Aug 2002) $
 *
 * Ksi - Interface to the Ksi intermediate language.
 */

type id;

/* Ksi type system. */
union ty {
	void Void;
	void Bool;
	int Int;		/* int is size */
	int Uint;
	int Float;
	ty Ptr;			/* ptr to ... */
	ty Reference;		/* reference to ... */
	*(ty, <ty>list) Func_ptr;
				/* (return_type, parameters) */
	void Ellipsis;		/* `...' to be used only in function parms */
	id Struct;
	id Union;
	id Macro_type;		/* Macro_type("c_int") -> .c_int */
	id Tyref;		/* (tyref foobar) */
	void Poly;
}

/* Given a string, return Ksi identifier.  */
id make_id(string);

/* Return type of signed integer of the same size as native pointer.  */
ty ptroff();

/* Ditto, but unsigned.  */
ty uptroff();

/* Attributes that associated with functions, variables and types.
   At the very moment only Extern, Local and Public are used.  However I
   hope to use No_return and Malloc_like some day...  */
union attr {
	int Aligned;
	void Packed;
	void Local		/* this is the default anyway... */;
	void Public;
	void Extern;
	void Suppress_debug;
	void Const_func;
	void Pure_func;
	void Malloc_like;
	void No_return;
	void Artificial;
	void Weak;
	void One_only;
	void Comdat;
	string Section;
	string Export_name;
	id Alias_set;
}

union binop {
	void Eq;		/* == */
	void Ne;		/* != */
	void Ge;		/* >= */
	void Gt;		/* > */
	void Le;		/* <= */
	void Lt;		/* < */
	void Plus;		/* + */
	void Minus;		/* - */
	void Mult;		/* * */
	void R_div;		/* float / float */
	void Exact_div;	/* division to use when no modulo is expected, like
			 * division of pointer difference by object size
			 * in C. */
	void Round_div;	/* round result */
	void Floor_div;	/* round result down */
	void Ceil_div;	/* round result up */
	void Trunc_div;	/* truncate result */

	/* corresponding modulus operations */
	void Round_mod;
	void Floor_mod;
	void Ceil_mod;
	void Trunc_mod;

	void Bit_and;	/* & */
	void Bit_ior;	/* | */
	void Bit_and_tc;	/* sth strange */
	void Bit_xor;	/* ^ */
	void Max;
	void Min;
	void L_rotate;
	void L_shift;	/* << */
	void R_rotate;
	void R_shift;	/* >> */

	/* ++/--, first operand is l-value to modify,
	 * second operand is how much to inc/decrement by. */
	void Post_inc;
	void Pre_inc;
	void Post_dec;
	void Pre_dec;

	void Truth_and_if;	/* && */
	void Truth_or_if;		/* || */
	void Truth_and;		/* similar to && but not lazy */
	void Truth_or;		/* similar to || but not lazy */
	void Truth_xor;		/* same, but ^ is used instead of | or & */
}

union unop {
	void Negate;		/* -X */
	void Bit_not;		/* ~X */
	void Truth_not;		/* !X */
	void Abs;
	void Addr;		/* &X */
	void R_addr;		/* As Addr, but return reference. */
	void Ffs;		/* position of first bit set */
	void Non_lvalue;		/* Hm... this is strange... */
}

union expr {
	/* (const ...) term */
	*(ty, Int64::t) Int_const;
	*(ty, float) Float_const;
	bool Bool_const;
	string String_const;

	id Enum_ref;

	/* null constant of type ty */
	ty Null;

	/* Single term of similar name: */
	*(ty, id) Var;
	*(ty, id) Static_var;
	expr Bind;		/* create scope */
	id Ref;			/* reference variable by name */
	id Func_ref;		/* get address of named function */
	*(expr, expr) Modify;	/* expr = expr */
	*(expr, id) Component;	/* expr.id */
	*(expr, id) Arrow;	/* expr->string, shortcut */

	/* this is shortcut to (return (modify (result) expr))
	 * expr can be Nop, (return (nil)) is generated then */
	expr Return;

	ty Size_of;			/* Returns (const .sizetype ...). */
	<expr>list Compound;		/* expr1, expr2, ..., exprn */
	expr Loop;			/* while (1) expr; */
	void Nop;
	void Result;
	expr Indirect;			/* *expr */
	*(expr, expr, expr) Cond;	/* expr ? expr : expr */
	/* call named function, passing parameters */
	*(id, <expr>list) Call;
	/* call function that is guaranteed not to have any side-effects */
	*(id, <expr>list) F_call;
	*(expr, <expr>list) Ptr_call;
	*(expr, <expr>list) Ptr_f_call;
	*(id, <expr>list) Built_in;
	*(id, <expr>list) F_built_in;
	*(ty, expr) Convert;
	*(id, expr) Block;		/* create named block */
	id Exit_block;			/* jump to end of named block */
	*(binop, expr, expr) Binary;
	*(unop, expr) Unary;
	*(expr, expr) Try_finally;		/* try e1 finally e2 */
	/* annotate expression with line number information. */
	*(int, expr) Line;
	/* annotate output Ksi code with comment */
	*(string, expr) Comment;
}

union decl {
	*(<attr>list, ty, id, <ty>list) Func_proto;
	*(<attr>list, ty, id, <*(ty, id)>list, expr) Func;
	string Top_comment;		/* output comment */
	id Struct_def_proto;
	*(<attr>list, id, <*(ty, id)>list) Struct_def;
	id Union_def_proto;
	*(<attr>list, id, <*(ty, id)>list) Union_def;
	*(<attr>list, id, ty) Ty_alias;
	*(<attr>list, id) Abstract_type;
	<*(id, int)>list Enum_def;
	*(<attr>list, ty, id) Global;
	string Input_file;
}

/* Print header of Ksi file to given channel.  */
void output_header(Stdio::FILE out);

/* Print one declaration to output channel.  */
void output_decl(Stdio::FILE out, decl d);

/* Finish writing Ksi file.  */
void output_trailer(Stdio::FILE out);

/* Set to true to get ugly-looking Ksi code (generation of it is however
   much faster).  Ergh... it doesn't work yet.  */
bool fast_writer;
