/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 975 $ $LastChangedDate: 2002-09-14 18:56:32 +0200 (Sat, 14 Sep 2002) $
 *
 * Gontlex - lexer for gont (but not only).
 */

open Gontlex;

struct <'a, 'b>mode {
    <'a>Hashtbl::t keywords;
    <'b>Hashtbl::t operators;
    Genlex::mode mode;
    bool keep_blanks;
    bool keep_comments;
    bool handle_tyvars;
}

struct <'a, 'b>t {
    <'a, 'b>mode m;
    Genlex::t genlex;
    *(int, int) beg_loc;
    *(int, int) end_loc;
    <Genlex::token>option saved_token;
    *(int, int) saved_loc;
    Genlex::token last_tok;
    *(int, int) last_loc;
    bool at_tok_beg;
}

def make_mode()
{
    _ r = {keywords = Hashtbl::create(0),
	   operators = Hashtbl::create(0),
	   keep_blanks = false,
	   keep_comments = false,
	   handle_tyvars = false,
	   mode = Genlex::make_mode()
	  };
    
    Genlex::add_operator(r.mode, "#");
    Genlex::add_operator(r.mode, "//");
    Genlex::add_operator(r.mode, "/*");
    Genlex::add_operator(r.mode, "\"");
    Genlex::add_operator(r.mode, "'");

    return r;
}

def of_genlex(m, gl)
{
    _ r = {genlex = gl,
	   beg_loc = (0, 0),
	   end_loc = (0, 0),
	   m = m,
	   at_tok_beg = true,
	   last_tok = Genlex::Eof,
	   last_loc = (0, 0),
	   saved_token = None,
	   saved_loc = (0, 0)
	};
    
    return r;
}

def beg_location(l) (l.beg_loc)
def end_location(l) (l.end_loc)
def add_keyword(l, s, v) (Hashtbl::add(l.keywords, s, v))

def add_operator(l, s, v)
{
    Genlex::add_operator(l.mode, s);
    Hashtbl::add(l.operators, s, v);
}

def keep_comments(l) {l.keep_comments = true;}
def keep_blanks(l) {l.keep_blanks = true;}
def handle_tyvars(l) {l.handle_tyvars = true;}

def get_gen_token(l)
{
    switch l.saved_token {
    case Some(t):
	l.last_loc = l.saved_loc;
	l.last_tok = t;
	l.saved_token = None;
    case None:
	l.last_loc = Genlex::get_location(l.genlex);
	l.last_tok = Genlex::get_token(l.genlex, l.m.mode);
    }
    
    if (l.at_tok_beg) {
	l.beg_loc = l.last_loc;
	l.at_tok_beg = false;
    }
//puts("ggt\n");
    
    return l.last_tok;
}

def unget_gen_token(l)
{
//puts("ugt\n");
    switch l.saved_token {
    case Some(_):
	Sys::abort();
	
    case None:
	l.saved_token = Some(l.last_tok);
	l.saved_loc = l.last_loc;
    }
}

def last_gen_token(l)
{
    l.at_tok_beg = true;
    
    switch l.saved_token {
    case Some(_):
	l.end_loc = l.saved_loc;
    case None:
	l.end_loc = Genlex::get_location(l.genlex);
    }
}

<char, char>Ptrhash::t escapes = Ptrhash::of_assoc_list([
    ('r', '\r'),
    ('n', '\n'),
    ('t', '\t'),
    ('e', '\e'),
    ('b', '\b'),
    ('"', '"'),
    ('\'', '\''),
    ('\\', '\\')
]);

def do_get_token(l)
{
    def parse_escape()
    {
	_ c = Genlex::peek_char(l.genlex);

	if (Hacks::same_chars(c, 'x')) {
	    Genlex::shift_char(l.genlex);
	    c = Genlex::peek_char(l.genlex);
	    
	    if (!Hacks::same_chars(c, '{')) 
		raise Error("\\x escape sequence doesn't start with {");
		
	    _ buf = String::to_char_array("0x");
	    
	    for (;;) {
		Genlex::shift_char(l.genlex);
		c = Genlex::peek_char(l.genlex);
		if (String::contains("\\\"'", c))
		    raise Error("run-away \\x{");
		if (Hacks::same_chars(c, '}'))
		    break;
		Array::append(buf, c);
		if (Array::length(buf) > 10)
		    raise Error("run-away \\x{");
	    }

	    Genlex::shift_char(l.genlex);

	    try {
		return char_of_int(atoi(String::of_char_array(buf)));
	    } with {
	    case Invalid_argument:
		raise Error("invalid \\x{...} seqeunce");
	    }
	} else {
	    Genlex::shift_char(l.genlex);
	    
	    try {
		return Ptrhash::find(escapes, c);
	    } with {
	    case Not_found:
		raise Error("unknown escape sequence");
	    }
	}
    }
   
    switch get_gen_token(l) {
    case Genlex::Blank(s):
//puts("blank ok\n");
	return Blank(s);
   
    case Genlex::Operator("\""):
	_ buf = Array::empty(100);
	string error = "";
	
	try {
	    for (;;) {
		_ c = Genlex::get_char(l.genlex);
		if (Hacks::same_chars(c, '\\'))
		    try {
			c = parse_escape();
		    } with {
		    case Error(e):
			if (error == "")
			    error = e;
		    }
		else if (Hacks::same_chars(c, '"'))
		    break;
		Array::append(buf, c);
	    }
	    
	    if (error != "")
		raise Error(error);

	    return String(String::of_char_array(buf));
	} with {
	case End_of_file:
	    if (error == "")
		error = "string literal doesn't end until EOF";
	    raise Error(error);
	}

    case Genlex::Operator("'"):
	string error = "";

	try {
	    char ret;
	    
	    ret = Genlex::get_char(l.genlex);

	    if (Hacks::same_chars(ret, '\\')) {
		try {
		   ret = parse_escape();
		} with {
		case Error(e):
		    if (error == "")
			error = e;
		}
	    } else {
		def id_char(c) (Unitype::is_ascii_alnum(c) || 
				Hacks::same_chars(c, '_'));
		_ buf = Array::empty(10);
		
		Array::append(buf, ret);
		try {
		    if (l.m.handle_tyvars && id_char(ret) && 
			!Hacks::same_chars(Genlex::peek_char(l.genlex), '\'')) {
			while (id_char(Genlex::peek_char(l.genlex)))
			    Array::append(buf, Genlex::get_char(l.genlex));
			return Ty_var(String::of_char_array(buf));
		    }
		} with {
		case End_of_file:
		    return Ty_var(String::of_char_array(buf));
		}
	    }

	    if (!Hacks::same_chars(Genlex::peek_char(l.genlex), '\'')) {
		if (error == "")
		    error = "excessive character";
		while (!String::contains("'\n", Genlex::get_char(l.genlex)))
		    skip;
	    } else {
		Genlex::shift_char(l.genlex);
	    }

	    if (error != "")
		raise Error(error);
	   
	    return Char(ret);
	} with {
	case End_of_file:
	    if (error == "")
		error = "character literal doesn't end until EOF";
	    raise Error(error);
	}
	    
    case Genlex::Operator("#" as o):
    case Genlex::Operator("//" as o):
	_ buf = String::to_char_array(o);
	try {
	    _ c;
	    do {
		c = Genlex::get_char(l.genlex);
		Array::append(buf, c);
	    } while (!Hacks::same_chars(c, '\n'));
	} with { case End_of_file: }
    //puts("comment ok\n");
	return Comment(String::of_char_array(buf));
    
    case Genlex::Operator("/*" as o):
	_ buf = String::to_char_array(o);
	_ lev = 1;
	bool seen_slash = false;
	bool seen_star = false;
	try {
	    while (lev > 0) {
		char c = Genlex::get_char(l.genlex);
		Array::append(buf, c);
		if (Hacks::same_chars(c, '/')) {
		    if (seen_star)
			lev--;
		    seen_slash = true;
		    seen_star = false;
		} else if (Hacks::same_chars(c, '*')) {
		    if (seen_slash)
			lev++;
		    seen_slash = false;
		    seen_star = true;
		} else {
		    seen_slash = false;
		    seen_star = false;
		}
	    }
	} with {
	case End_of_file:
	    raise Error("run-away comment");
	}
	
	return Comment(String::of_char_array(buf));

    case Genlex::Operator(o):
	if (Hashtbl::mem(l.m.operators, o))
	    return Operator(Hashtbl::find(l.m.operators, o));
	else
	    raise Error("invalid operator in input `" + o + "'");
	
    case Genlex::Id(i):
	if (Hashtbl::mem(l.m.keywords, i))
	    return Keyword(Hashtbl::find(l.m.keywords, i));
	else
	    return Id(i);

    case Genlex::Number(n):
	try {
	    if (String::contains(n, '.') || 
		String::contains(n, 'e') ||
		String::contains(n, 'E')) {
		ignore(atof(n));
		return Float(n);
	    } else {
		// TODO: if we'll have bignum, use it here
		ignore(Int64::of_string(n));
		return Int(n);
	    }
	} with {
	case Invalid_argument(_):
	    raise Error("invalid number `" + n + "'");
	}
    
    case Genlex::Eof:
	return Eof;

    case Genlex::Junk(c):
	// send junk in one piece
	_ buf = Array::empty(100);
	Array::append(buf, c);
	loop: for (;;)
	    switch get_gen_token(l) {
	    case Genlex::Junk(c):
		Array::append(buf, c);
	    case _:
		unget_gen_token(l);
		break loop;
	    }
	raise Error("junk in input");
    }
}

def get_token(l)
{
    try {
	for (;;) {
//puts("loop\n");
	    _ t = do_get_token(l);
	    last_gen_token(l);
	    switch t {
	    case Comment(_) if !l.m.keep_comments:
	    case Blank(_) if !l.m.keep_blanks:
		skip;
	    case _:
		return t;
	    }
	}
    } with {
    case Error(_) as exn:
	last_gen_token(l);
	raise exn;
    }
}

// vim:sts=4:sw=4
